﻿/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

using System;
using System.Collections.Generic;
using System.IO;

namespace Npoi.Core.POIFS.FileSystem
{
    /// <summary>
    /// A DirectoryEntry filter, which exposes another  DirectoryEntry less certain parts.
    /// This is typically used when copying or comparing  Filesystems.
    /// </summary>
    public class FilteringDirectoryNode : DirectoryEntry
    {
        private List<string> excludes;
        private Dictionary<string, List<string>> childExcludes;
        private DirectoryEntry directory;

        /// <summary>
        /// Creates a filter round the specified directory, which will exclude entries such as
        /// "MyNode" and "MyDir/IgnoreNode". The excludes can stretch into children, if they contain a /.
        /// </summary>
        /// <param name="directory">The Directory to filter</param>
        /// <param name="excludes">The Entries to exclude</param>
        public FilteringDirectoryNode(DirectoryEntry directory, ICollection<string> excludes)
        {
            this.directory = directory;

            // Process the excludes
            this.excludes = new List<string>();
            childExcludes = new Dictionary<string, List<string>>();
            foreach (string excl in excludes)
            {
                int splitAt = excl.IndexOf('/');
                if (splitAt == -1)
                {
                    // Applies to us
                    this.excludes.Add(excl);
                }
                else
                {
                    // Applies to a child
                    string child = excl.Substring(0, splitAt);
                    string childExcl = excl.Substring(splitAt + 1);
                    if (!childExcludes.ContainsKey(child))
                    {
                        childExcludes.Add(child, new List<string>());
                    }
                    childExcludes[child].Add(childExcl);
                }
            }
        }

        #region DirectoryEntry 成员

        public IEnumerator<Entry> Entries
        {
            get { return GetEntries(); }
        }

        public List<string> EntryNames
        {
            get
            {
                List<string> names = new List<string>();
                foreach (string name in directory.EntryNames)
                {
                    if (!excludes.Contains(name))
                    {
                        names.Add(name);
                    }
                }
                return names;
            }
        }

        public bool IsEmpty
        {
            get { return EntryCount == 0; }
        }

        public bool HasEntry(string name)
        {
            if (excludes.Contains(name))
            {
                return false;
            }
            return directory.HasEntry(name);
        }

        public int EntryCount
        {
            get
            {
                int size = directory.EntryCount;
                foreach (string excl in excludes)
                {
                    if (directory.HasEntry(excl))
                    {
                        size--;
                    }
                }
                return size;
            }
        }

        public IEnumerator<Entry> GetEntries()
        {
            return new FilteringIterator(this); ;
        }

        public Entry GetEntry(string name)
        {
            if (excludes.Contains(name))
            {
                throw new FileNotFoundException(name);
            }

            Entry entry = directory.GetEntry(name);
            return WrapEntry(entry);
        }

        private Entry WrapEntry(Entry entry)
        {
            string name = entry.Name;
            if (childExcludes.ContainsKey(name) && entry is DirectoryEntry)
            {
                return new FilteringDirectoryNode(
                      (DirectoryEntry)entry, childExcludes[name]);
            }
            return entry;
        }

        public DocumentEntry CreateDocument(string name, System.IO.Stream stream)
        {
            return directory.CreateDocument(name, stream);
        }

        public DocumentEntry CreateDocument(string name, int size, EventFileSystem.POIFSWriterListener writer)
        {
            return directory.CreateDocument(name, size, writer);
        }

        public DirectoryEntry CreateDirectory(string name)
        {
            return directory.CreateDirectory(name);
        }

        public Util.ClassID StorageClsid
        {
            get
            {
                return directory.StorageClsid;
            }
            set
            {
                directory.StorageClsid = value;
            }
        }

        #endregion DirectoryEntry 成员

        #region Entry 成员

        public string Name
        {
            get { return directory.Name; }
        }

        public bool IsDirectoryEntry
        {
            get { return true; }
        }

        public bool IsDocumentEntry
        {
            get { return false; }
        }

        public DirectoryEntry Parent
        {
            get { return directory.Parent; }
        }

        public bool Delete()
        {
            return directory.Delete();
        }

        public bool RenameTo(string newName)
        {
            return directory.RenameTo(newName);
        }

        #endregion Entry 成员

        #region IEnumerable<Entry> 成员

        public IEnumerator<Entry> GetEnumerator()
        {
            return new FilteringIterator(this);
        }

        #endregion IEnumerable<Entry> 成员

        #region IEnumerable 成员

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return new FilteringIterator(this);
        }

        #endregion IEnumerable 成员

        private class FilteringIterator : IEnumerator<Entry>
        {
            private IEnumerator<Entry> parent;
            private Entry next;
            private DirectoryEntry directory;
            private FilteringDirectoryNode filtering;

            public FilteringIterator(FilteringDirectoryNode filtering)
            {
                this.filtering = filtering;
                directory = filtering.directory;
                parent = directory.Entries;
                //LocateNext();
            }

            //private void LocateNext()
            //{
            //    next = null;
            //    Entry e;
            //    while (parent.MoveNext() && next == null)
            //    {
            //        e = parent.Current;
            //        if (!filtering.excludes.Contains(e.Name))
            //        {
            //            next = filtering.WrapEntry(e);
            //        }
            //    }
            //}

            //public bool HasNext()
            //{
            //    return (next != null);
            //}

            //public Entry Next()
            //{
            //    Entry e = next;
            //    LocateNext();
            //    return e;
            //}

            public void Remove()
            {
                throw new InvalidOperationException("Remove not supported");
            }

            #region IEnumerator<Entry> 成员

            public Entry Current
            {
                get { return next; }
            }

            #endregion IEnumerator<Entry> 成员

            #region IDisposable 成员

            public void Dispose()
            {
            }

            #endregion IDisposable 成员

            #region IEnumerator 成员

            object System.Collections.IEnumerator.Current
            {
                get { return next; }
            }

            public bool MoveNext()
            {
                next = null;
                Entry e;
                while (parent.MoveNext())
                {
                    e = parent.Current;
                    if (!filtering.excludes.Contains(e.Name))
                    {
                        next = filtering.WrapEntry(e);
                        break;
                    }
                }
                return (next != null);
                //throw new NotImplementedException();
            }

            public void Reset()
            {
                throw new NotImplementedException();
            }

            #endregion IEnumerator 成员
        }
    }
}